package PDF::Builder::Resource::Shading;

use base 'PDF::Builder::Resource';

use strict;
use warnings;

our $VERSION = '3.014'; # VERSION
my $LAST_UPDATE = '2.029'; # manually update whenever code is changed

=head1 NAME

PDF::Builder::Resource::Shading - support stub for shading patterns. Inherits from L<PDF::Builder::Resource>

=cut

1;
